# -*- coding: iso-8859-1 -*-
"""
    MoinMoin - EventAggregator library

    @copyright: 2008, 2009, 2010 by Paul Boddie <paul@boddie.org.uk>
    @copyright: 2000-2004 Juergen Hermann <jh@web.de>,
                2005-2008 MoinMoin:ThomasWaldmann.
    @license: GNU GPL (v2 or later), see COPYING.txt for details.
"""

from MoinMoin.Page import Page
from MoinMoin import search, version
from MoinMoin import wikiutil
import calendar
import datetime
import time
import re

__version__ = "0.5"

# Date labels.

month_labels = ["January", "February", "March", "April", "May", "June",
    "July", "August", "September", "October", "November", "December"]
weekday_labels = ["Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun"]

# Regular expressions where MoinMoin does not provide the required support.

category_regexp = None

# Page parsing.

definition_list_regexp = re.compile(ur'(?P<wholeterm>^(?P<optcomment>#*)\s+(?P<term>.*?)::\s)(?P<desc>.*?)$', re.UNICODE | re.MULTILINE)
category_membership_regexp = re.compile(ur"^\s*((Category\S+)(\s+Category\S+)*)\s*$", re.MULTILINE | re.UNICODE)

# Value parsing.

date_regexp = re.compile(ur'(?P<year>[0-9]{4})-(?P<month>[0-9]{2})-(?P<day>[0-9]{2})', re.UNICODE)
month_regexp = re.compile(ur'(?P<year>[0-9]{4})-(?P<month>[0-9]{2})', re.UNICODE)
verbatim_regexp = re.compile(ur'(?:'
    ur'<<Verbatim\((?P<verbatim>.*?)\)>>'
    ur'|'
    ur'\[\[Verbatim\((?P<verbatim2>.*?)\)\]\]'
    ur'|'
    ur'`(?P<monospace>.*?)`'
    ur'|'
    ur'{{{(?P<preformatted>.*?)}}}'
    ur')', re.UNICODE)

# Utility functions.

def isMoin15():
    return version.release.startswith("1.5.")

def getCategoryPattern(request):
    global category_regexp

    try:
        return request.cfg.cache.page_category_regexact
    except AttributeError:

        # Use regular expression from MoinMoin 1.7.1 otherwise.

        if category_regexp is None:
            category_regexp = re.compile(u'^%s$' % ur'(?P<all>Category(?P<key>(?!Template)\S+))', re.UNICODE)
        return category_regexp

# Textual representations.

def getHTTPTimeString(tmtuple):
    return "%s, %02d %s %04d %02d:%02d:%02d GMT" % (
        weekday_labels[tmtuple.tm_wday],
        tmtuple.tm_mday,
        month_labels[tmtuple.tm_mon -1], # zero-based labels
        tmtuple.tm_year,
        tmtuple.tm_hour,
        tmtuple.tm_min,
        tmtuple.tm_sec
        )

def getSimpleWikiText(text):

    """
    Return the plain text representation of the given 'text' which may employ
    certain Wiki syntax features, such as those providing verbatim or monospaced
    text.
    """

    # NOTE: Re-implementing support for verbatim text and linking avoidance.

    return "".join([s for s in verbatim_regexp.split(text) if s is not None])

def getEncodedWikiText(text):

    "Encode the given 'text' in a verbatim representation."

    return "<<Verbatim(%s)>>" % text

def getPrettyTitle(title):

    "Return a nicely formatted version of the given 'title'."

    return title.replace("_", " ").replace("/", u"  ")

def getMonthLabel(month):

    "Return an unlocalised label for the given 'month'."

    return month_labels[month - 1] # zero-based labels

def getDayLabel(weekday):

    "Return an unlocalised label for the given 'weekday'."

    return weekday_labels[weekday]

# Action support functions.

def getPageRevision(page):

    "Return the revision details dictionary for the given 'page'."

    # From Page.edit_info...

    if hasattr(page, "editlog_entry"):
        line = page.editlog_entry()
    else:
        line = page._last_edited(page.request) # MoinMoin 1.5.x and 1.6.x

    timestamp = line.ed_time_usecs
    mtime = wikiutil.version2timestamp(long(timestamp)) # must be long for py 2.2.x
    return {"timestamp" : time.gmtime(mtime), "comment" : line.comment}

# Category discovery and searching.

def getCategories(request):

    """
    From the AdvancedSearch macro, return a list of category page names using
    the given 'request'.
    """

    # This will return all pages with "Category" in the title.

    cat_filter = getCategoryPattern(request).search
    return request.rootpage.getPageList(filter=cat_filter)

def getCategoryMapping(category_pagenames, request):

    """
    For the given 'category_pagenames' return a list of tuples of the form
    (category name, category page name) using the given 'request'.
    """

    cat_pattern = getCategoryPattern(request)
    mapping = []
    for pagename in category_pagenames:
        name = cat_pattern.match(pagename).group("key")
        if name != "Category":
            mapping.append((name, pagename))
    mapping.sort()
    return mapping

def getCategoryPages(pagename, request):

    """
    Return the pages associated with the given category 'pagename' using the
    'request'.
    """

    query = search.QueryParser().parse_query('category:%s' % pagename)
    if isMoin15():
        results = search.searchPages(request, query)
        results.sortByPagename()
    else:
        results = search.searchPages(request, query, "page_name")

    cat_pattern = getCategoryPattern(request)
    pages = []
    for page in results.hits:
        if not cat_pattern.match(page.page_name):
            pages.append(page)
    return pages

# The main activity functions.

class EventPage:

    "An event page."

    def __init__(self, page):
        self.page = page
        self.details = None
        self.body = None
        self.categories = None

    def __cmp__(self, other):

        """
        Compare this object with 'other' using the event start and end details.
        """

        event_details1 = self.getEventDetails()
        event_details2 = other.getEventDetails()
        return cmp(
            (event_details1["start"], event_details1["end"]),
            (event_details2["start"], event_details2["end"])
            )

    def copyPage(self, page):

        "Copy the body of the given 'page'."

        self.body = page.getBody()

    def getPageURL(self, request):

        "Using 'request', return the URL of this page."

        page = self.page

        if isMoin15():
            return request.getQualifiedURL(page.url(request))
        else:
            return request.getQualifiedURL(page.url(request, relative=0))

    def getFormat(self):

        "Get the format used on this page."

        if isMoin15():
            return "wiki" # page.pi_format
        else:
            return self.page.pi["format"]

    def getRevisions(self):

        "Return a list of page revisions."

        return self.page.getRevList()

    def getPageRevision(self):

        "Return the revision details dictionary for this page."

        return getPageRevision(self.page)

    def getPageName(self):

        "Return the page name."

        return self.page.page_name

    def getPrettyPageName(self):

        "Return a nicely formatted title/name for this page."

        return getPrettyPageName(self.page)

    def getBody(self):

        "Get the current page body."

        if self.body is None:
            self.body = self.page.get_raw_body()
        return self.body

    def getEventDetails(self):

        "Return a dictionary of event details from this page."

        if self.details is None:
            self.details = {}

            if self.getFormat() == "wiki":
                for match in definition_list_regexp.finditer(self.getBody()):

                    # Skip commented-out items.

                    if match.group("optcomment"):
                        continue

                    # Permit case-insensitive list terms.

                    term = match.group("term").lower()
                    desc = match.group("desc")

                    # Special value type handling.

                    # Dates.

                    if term in ("start", "end"):
                        desc = getDate(desc)

                    # Lists (whose elements may be quoted).

                    elif term in ("topics", "categories"):
                        desc = [getSimpleWikiText(value.strip()) for value in desc.split(",")]

                    # Labels which may well be quoted.

                    elif term in ("title", "summary", "description"):
                        desc = getSimpleWikiText(desc)

                    if desc is not None:
                        self.details[term] = desc

        return self.details

    def getCategoryMembership(self):

        "Get the category names from this page."

        if self.categories is None:
            body = self.getBody()
            match = category_membership_regexp.search(body)
            self.categories = match.findall().split()

        return self.categories

    def getEventSummary(self, event_parent=None):

        """
        Return either the given title or summary of the event described by this
        page, according to the page's event details, or using the pretty version
        of the page name.

        If the optional 'event_parent' is specified, any page beneath the given
        'event_parent' page in the page hierarchy will omit this parent information
        if its name is used as the summary.
        """

        event_details = self.getEventDetails()

        if event_details.has_key("title"):
            return event_details["title"]
        elif event_details.has_key("summary"):
            return event_details["summary"]
        else:
            # If appropriate, remove the parent details and "/" character.

            title = self.getPageName()

            if event_parent is not None and title.startswith(event_parent):
                title = title[len(event_parent.rstrip("/")) + 1:]

            return getPrettyTitle(title)

    def setEventDetails(self, event_details):

        "Set the 'event_details' for this page."

        self.details = event_details

    def setCategoryMembership(self, category_names):

        """
        Set the category membership for the page using the specified
        'category_names'.
        """

        self.categories = category_names

    def flushEventDetails(self):

        "Flush the current event details to this page's body text."

        new_body_parts = []
        end_of_last_match = 0
        body = self.getBody()
        event_details = self.getEventDetails()

        for match in definition_list_regexp.finditer(body):

            # Add preceding text to the new body.

            new_body_parts.append(body[end_of_last_match:match.start()])
            end_of_last_match = match.end()

            # Get the matching regions, adding the term to the new body.

            new_body_parts.append(match.group("wholeterm"))

            # Permit case-insensitive list terms.

            term = match.group("term").lower()
            desc = match.group("desc")

            # Special value type handling.

            if event_details.has_key(term):

                # Dates.

                if term in ("start", "end"):
                    desc = desc.replace("YYYY-MM-DD", str(event_details[term]))

                # Lists (whose elements may be quoted).

                elif term in ("topics", "categories"):
                    desc = ", ".join(getEncodedWikiText(event_details[term]))

                # Labels which may well be quoted.

                elif term in ("title", "summary"):
                    desc = getEncodedWikiText(event_details[term])

                # Text which need not be quoted, but it will be Wiki text.

                elif term in ("description",):
                    desc = event_details[term]

            new_body_parts.append(desc)

        else:
            new_body_parts.append(body[end_of_last_match:])

        self.body = "".join(new_body_parts)

    def flushCategoryMembership(self):

        "Flush the category membership to the page body."

        body = self.getBody()
        category_names = self.getCategoryMembership()
        match = category_membership_regexp.search(body)

        if match:
            self.body = "".join([body[:match.start()], " ".join(category_names), body[match.end():]])

    def saveChanges(self):

        "Save changes to the event."

        self.flushEventDetails()
        self.flushCategoryMembership()
        self.page.saveText(self.getBody(), 0)

    def linkToPage(self, request, text, query_string=None):

        """
        Using 'request', return a link to this page with the given link 'text'
        and optional 'query_string'.
        """

        return linkToPage(request, self.page, text, query_string)

def getEvents(request, category_names, calendar_start=None, calendar_end=None):

    """
    Using the 'request', generate a list of events found on pages belonging to
    the specified 'category_names', using the optional 'calendar_start' and
    'calendar_end' month tuples of the form (year, month) to indicate a window
    of interest.

    Return a list of events, a dictionary mapping months to event lists (within
    the window of interest), a list of all events within the window of interest,
    the earliest month of an event within the window of interest, and the latest
    month of an event within the window of interest.
    """

    # Re-order the window, if appropriate.

    if calendar_start is not None and calendar_end is not None and calendar_start > calendar_end:
        calendar_start, calendar_end = calendar_end, calendar_start

    events = []
    shown_events = {}
    all_shown_events = []
    processed_pages = set()

    earliest = None
    latest = None

    for category_name in category_names:

        # Get the pages and page names in the category.

        pages_in_category = getCategoryPages(category_name, request)

        # Visit each page in the category.

        for page_in_category in pages_in_category:
            pagename = page_in_category.page_name

            # Only process each page once.

            if pagename in processed_pages:
                continue
            else:
                processed_pages.add(pagename)

            # Get a real page, not a result page.

            event_page = EventPage(Page(request, pagename))
            event_details = event_page.getEventDetails()

            # Remember the event page.

            events.append(event_page)

            # Test for the suitability of the event.

            if event_details.has_key("start") and event_details.has_key("end"):

                start_month = event_details["start"].as_month()
                end_month = event_details["end"].as_month()

                # Compare the months of the dates to the requested calendar
                # window, if any.

                if (calendar_start is None or end_month >= calendar_start) and \
                    (calendar_end is None or start_month <= calendar_end):

                    all_shown_events.append(event_page)

                    if earliest is None or start_month < earliest:
                        earliest = start_month
                    if latest is None or end_month > latest:
                        latest = end_month

                    # Store the event in the month-specific dictionary.

                    first = max(start_month, calendar_start or start_month)
                    last = min(end_month, calendar_end or end_month)

                    for event_month in first.months_until(last):
                        if not shown_events.has_key(event_month):
                            shown_events[event_month] = []
                        shown_events[event_month].append(event_page)

    return events, shown_events, all_shown_events, earliest, latest

def setEventTimestamps(request, events):

    """
    Using 'request', set timestamp details in the details dictionary of each of
    the 'events'.

    Retutn the latest timestamp found.
    """

    latest = None

    for event_page in events:
        event_details = event_page.getEventDetails()

        # Get the initial revision of the page.

        revisions = event_page.getRevisions()
        event_page_initial = Page(request, event_page.getPageName(), rev=revisions[-1])

        # Get the created and last modified times.

        initial_revision = getPageRevision(event_page_initial)
        event_details["created"] = initial_revision["timestamp"]
        latest_revision = event_page.getPageRevision()
        event_details["last-modified"] = latest_revision["timestamp"]
        event_details["sequence"] = len(revisions) - 1
        event_details["last-comment"] = latest_revision["comment"]

        if latest is None or latest < event_details["last-modified"]:
            latest = event_details["last-modified"]

    return latest

def getOrderedEvents(events):

    """
    Return a list with the given 'events' ordered according to their start and
    end dates.
    """

    ordered_events = events[:]
    ordered_events.sort()
    return ordered_events

def getConcretePeriod(calendar_start, calendar_end, earliest, latest):

    """
    From the requested 'calendar_start' and 'calendar_end', which may be None,
    indicating that no restriction is imposed on the period for each of the
    boundaries, use the 'earliest' and 'latest' event months to define a
    specific period of interest.
    """

    # Define the period as starting with any specified start month or the
    # earliest event known, ending with any specified end month or the latest
    # event known.

    first = calendar_start or earliest
    last = calendar_end or latest

    # If there is no range of months to show, perhaps because there are no
    # events in the requested period, and there was no start or end month
    # specified, show only the month indicated by the start or end of the
    # requested period. If all events were to be shown but none were found show
    # the current month.

    if first is None:
        first = last or getCurrentMonth()
    if last is None:
        last = first or getCurrentMonth()

    # Permit "expiring" periods (where the start date approaches the end date).

    return min(first, last), last

def getCoverage(start, end, events):

    """
    Within the period defined by the 'start' and 'end' dates, determine the
    coverage of the days in the period by the given 'events', returning a set of
    covered days, along with a list of slots, where each slot contains a tuple
    of the form (set of covered days, events).
    """

    all_events = []
    full_coverage = set()

    # Get event details.

    for event_page in events:
        event_details = event_page.getEventDetails()

        # Test for the event in the period.

        if event_details["start"] <= end and event_details["end"] >= start:

            # Find the coverage of this period for the event.

            event_start = max(event_details["start"], start)
            event_end = min(event_details["end"], end)
            event_coverage = set(event_start.days_until(event_end))

            # Update the overall coverage.

            full_coverage.update(event_coverage)

            # Try and fit the event into the events list.

            for i, (coverage, covered_events) in enumerate(all_events):

                # Where the event does not overlap with the current
                # element, add it alongside existing events.

                if not coverage.intersection(event_coverage):
                    covered_events.append(event_page)
                    all_events[i] = coverage.union(event_coverage), covered_events
                    break

            # Make a new element in the list if the event cannot be
            # marked alongside existing events.

            else:
                all_events.append((event_coverage, [event_page]))

    return full_coverage, all_events

# Date-related functions.

class Period:

    "A simple period of time."

    def __init__(self, data):
        self.data = data

    def months(self):
        return self.data[0] * 12 + self.data[1]

class Month:

    "A simple year-month representation."

    def __init__(self, data):
        self.data = tuple(data)

    def __repr__(self):
        return "%s(%r)" % (self.__class__.__name__, self.data)

    def __str__(self):
        return "%04d-%02d" % self.as_tuple()[:2]

    def __hash__(self):
        return hash(self.as_tuple())

    def as_tuple(self):
        return self.data

    def as_date(self, day):
        return Date(self.as_tuple() + (day,))

    def year(self):
        return self.data[0]

    def month(self):
        return self.data[1]

    def month_properties(self):

        """
        Return the weekday of the 1st of the month, along with the number of
        days, as a tuple.
        """

        year, month = self.data
        return calendar.monthrange(year, month)

    def month_update(self, n=1):

        "Return the month updated by 'n' months."

        year, month = self.data
        return Month((year + (month - 1 + n) / 12, (month - 1 + n) % 12 + 1))

    def next_month(self):

        "Return the month following this one."

        return self.month_update(1)

    def previous_month(self):

        "Return the month preceding this one."

        return self.month_update(-1)

    def __sub__(self, start):

        """
        Return the difference in years and months between this month and the
        'start' month as a period.
        """

        return Period([(x - y) for x, y in zip(self.data, start.data)])

    def __cmp__(self, other):
        return cmp(self.data, other.data)

    def until(self, end, nextfn, prevfn):
        month = self
        months = [month]
        if month < end:
            while month < end:
                month = nextfn(month)
                months.append(month)
        elif month > end:
            while month > end:
                month = prevfn(month)
                months.append(month)
        return months

    def months_until(self, end):
        return self.until(end, Month.next_month, Month.previous_month)

class Date(Month):

    "A simple year-month-day representation."

    def __str__(self):
        return "%04d-%02d-%02d" % self.as_tuple()[:3]

    def as_month(self):
        return Month(self.data[:2])

    def day(self):
        return self.data[2]

    def next_day(self):

        "Return the date following this one."

        year, month, day = self.data
        _wd, end_day = calendar.monthrange(year, month)
        if day == end_day:
            if month == 12:
                return Date((year + 1, 1, 1))
            else:
                return Date((year, month + 1, 1))
        else:
            return Date((year, month, day + 1))

    def previous_day(self):

        "Return the date preceding this one."

        year, month, day = self.data
        if day == 1:
            if month == 1:
                return Date((year - 1, 12, 31))
            else:
                _wd, end_day = calendar.monthrange(year, month - 1)
                return Date((year, month - 1, end_day))
        else:
            return Date((year, month, day - 1))

    def days_until(self, end):
        return self.until(end, Date.next_day, Date.previous_day)

def getDate(s):

    "Parse the string 's', extracting and returning a date string."

    m = date_regexp.search(s)
    if m:
        return Date(map(int, m.groups()))
    else:
        return None

def getMonth(s):

    "Parse the string 's', extracting and returning a month string."

    m = month_regexp.search(s)
    if m:
        return Month(map(int, m.groups()))
    else:
        return None

def getCurrentMonth():

    "Return the current month as a (year, month) tuple."

    today = datetime.date.today()
    return Month((today.year, today.month))

def getCurrentYear():

    "Return the current year."

    today = datetime.date.today()
    return today.year

# User interface functions.

def getParameter(request, name, default=None):
    return request.form.get(name, [default])[0]

def getQualifiedParameter(request, calendar_name, argname, default=None):
    argname = getQualifiedParameterName(calendar_name, argname)
    return getParameter(request, argname, default)

def getQualifiedParameterName(calendar_name, argname):
    if calendar_name is None:
        return argname
    else:
        return "%s-%s" % (calendar_name, argname)

def getParameterMonth(arg):

    "Interpret 'arg', recognising keywords and simple arithmetic operations."

    n = None

    if arg.startswith("current"):
        date = getCurrentMonth()
        if len(arg) > 8:
            n = int(arg[7:])

    elif arg.startswith("yearstart"):
        date = Month((getCurrentYear(), 1))
        if len(arg) > 10:
            n = int(arg[9:])

    elif arg.startswith("yearend"):
        date = Month((getCurrentYear(), 12))
        if len(arg) > 8:
            n = int(arg[7:])

    else:
        date = getMonth(arg)

    if n is not None:
        date = date.month_update(n)

    return date

def getFormMonth(request, calendar_name, argname):

    """
    Return the month from the 'request' for the calendar with the given
    'calendar_name' using the parameter having the given 'argname'.
    """

    arg = getQualifiedParameter(request, calendar_name, argname)
    if arg is not None:
        return getParameterMonth(arg)
    else:
        return None

def getFormMonthPair(request, yeararg, montharg):

    """
    Return the month from the 'request' for the calendar with the given
    'calendar_name' using the parameters having the given 'yeararg' and
    'montharg' names.
    """

    year = getParameter(request, yeararg)
    month = getParameter(request, montharg)
    if year and month:
        return Month((int(year), int(month)))
    else:
        return None

# Page-related functions.

def getPrettyPageName(page):

    "Return a nicely formatted title/name for the given 'page'."

    if isMoin15():
        title = page.split_title(page.request, force=1)
    else:
        title = page.split_title(force=1)

    return getPrettyTitle(title)

def linkToPage(request, page, text, query_string=None):

    """
    Using 'request', return a link to 'page' with the given link 'text' and
    optional 'query_string'.
    """

    text = wikiutil.escape(text)

    if isMoin15():
        url = wikiutil.quoteWikinameURL(page.page_name)
        if query_string is not None:
            url = "%s?%s" % (url, query_string)
        return wikiutil.link_tag(request, url, text, getattr(page, "formatter", None))
    else:
        return page.link_to_raw(request, text, query_string)

# vim: tabstop=4 expandtab shiftwidth=4
